// -*- C++ -*-
/*!
 * @file  rtc_3DShape.h
 * @brief Processing of three dimensional image creation
 * @date  $Date$
 *
 * @author EAST JAPAN ACCOUNTING CENTER CO.,LTD.
 *
 * BSD 3-clause "New" or "Revised" License
 *
 * $Id$
 */

#ifndef RTC_3DSHAPE_H
#define RTC_3DSHAPE_H

#include <rtm/idl/BasicDataTypeSkel.h>
#include <rtm/idl/ExtendedDataTypesSkel.h>
#include <rtm/idl/InterfaceDataTypesSkel.h>

// add to code
#include <iostream>
#include <string>
#include <stdio.h>
#include <direct.h>
#include <Shlwapi.h>
#include <windows.h>
#include <math.h>

// add to code for OpenCV
#include <opencv2/core/core.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>

#include "local_basic_type.h"
#include "3dshape_service.h"
#include "2DCDP/2DCDP4.h"
#include "2DCDP/2DCDP4FTV.h"
#include "FactorizationController/factorizationcontroller.h"
#include "CDPObjectFormatter/CDPObjectFormatter.h"
#include "Object/ObjectMesh.h"
#include "Creater/Creater.h"

// Service implementation headers
// <rtc-template block="service_impl_h">
#include "Creation3DResultStatusSVC_impl.h"

// </rtc-template>

// Service Consumer stub headers
// <rtc-template block="consumer_stub_h">
#include "RltObjectIFStub.h"
#include "Rlt3DShapeIFStub.h"

// </rtc-template>

// Service Consumer stub headers
// <rtc-template block="port_stub_h">
using namespace RltObjectIF;
using namespace Rlt3DShapeIF;
// </rtc-template>

using namespace RTC;

#include <rtm/Manager.h>
#include <rtm/DataFlowComponentBase.h>
#include <rtm/CorbaPort.h>
#include <rtm/DataInPort.h>
#include <rtm/DataOutPort.h>

// add to code
/*=================================================================================================
  Macro definition
 ================================================================================================*/
//#define _DEBUG_SW_

//#define RESULT_OK                    (1)
//#define RESULT_NG                    (0)

#define _3DSHAPE_STAGE_INIT            (0)
#define _3DSHAPE_STAGE_MIN             (_3DSHAPE_STAGE_INIT)
#define _3DSHAPE_STAGE1                (1)
#define _3DSHAPE_STAGE2                (2)
#define _3DSHAPE_STAGE3                (3)
#define ERROR_DETECTED                 _3DSHAPE_STAGE3
#define _3DSHAPE_STAGE_MAX             (_3DSHAPE_STAGE3+1)

#define IMAGES_LENGTH_2                (2)
#define FORMAT_COLOR24                 (24)
#define CHANNEL_NUM                    (3)

#define XIMG_120                       (120)
#define YIMG_90                        (90)

#define XIMG_320                       (320)
#define YIMG_240                       (240)

#define _ON_INITIALIZE                 ("onInitialize execution")
#define _ON_ACTIVATED                  ("onActivated execution")
#define _ON_DEACTIVATED                ("onDeactivated execution")
#define _ON_ABORTING                   ("onAborting execution")
#define _ON_RESET                      ("onReset execution")
#define _ERROR_RESET                   ("ERROR: Error was detected. Please reset the 3D shape RTC by the RT System Editor.")

#define _ERROR_IMAGE_COLOR             ("ERROR: This image data is not implemented in this program")
#define _ERROR_IMAGE_EMPTY             ("ERROR: Empty error of image data")
#define _ERROR_IMAGE_SIZE              ("ERROR: This image data is outside the size target")
#define _STAGE_ERROR                   ("ERROR: Wrong value of the stage variable")
#define _ERROR_WRITE_OBJ               ("ERROR: ResultObject data port is not connected. please check the connect of data port!")
#define _ERROR_WRITE_IMG               ("ERROR: Image data port is not connected. please check the connect of data port!")
#define _ERROR_WRITE_PATH              ("ERROR: Path information port is not connected. please check the connect of data port!")
#define _ERROR_OPEN_FILE               ("ERROR: Failed to open the 3D shape file!")
#define _ERROR_ALLOCATION              ("ERROR: Memory allocation error")
#define _ERROR_HEAP_MEM                ("ERROR: Can't heap memory")
#define _ERROR_HEAP_MEM_OBJ            ("ERROR: Z coordinate creation failed")

#define _3D_SHAPE_SUCCESS              ("3D shape RTC process was successful")
#define _3D_SHAPE_FAILED               ("3D shape RTC process was failed")

#define DEPTH_POWER                    50

#define IMG_POS_C                      (0)
#define IMG_POS_L                      (1)
#define IMG_POS_R                      IMG_POS_L
#define IMG_INP_C                      (1)

#define DEF_MAX_LENGTH                 (1024)
#define DEF_MID_LENGTH                 (512)
#define DEF_MIN_LENGTH                 (64)

#define DEF_GET_PATH_NG                (0)
#define DEF_GET_PATH_OK                (1)

#define DEF_NUM_INIT                   (0)

/*!
 * @class rtc_3DShape
 * @brief Processing of three dimensional image creation
 *
 * Processing of three dimensional image creation
 *
 * 3D image object and cameraimage
 *
 * -
 *
 * http://opencv.org/
 *
 */
class rtc_3DShape
  : public RTC::DataFlowComponentBase
{
 public:
  /*!
   * @brief constructor
   * @param manager Maneger Object
   */
  rtc_3DShape(RTC::Manager* manager);

  /*!
   * @brief destructor
   */
  ~rtc_3DShape();

  // <rtc-template block="public_attribute">
  
  // </rtc-template>

  // <rtc-template block="public_operation">
  
  // </rtc-template>

  /***
   * Initialization of the 3D Shape RTC
   *
   * The initialize action (on CREATED->ALIVE transition)
   * formaer rtc_init_entry() 
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre -
   * @post -
   * 
   */
   virtual RTC::ReturnCode_t onInitialize();

  /***
   * 
   *
   * The finalize action (on ALIVE->END transition)
   * formaer rtc_exiting_entry()
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onFinalize();

  /***
   *
   * The startup action when ExecutionContext startup
   * former rtc_starting_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onStartup(RTC::UniqueId ec_id);

  /***
   *
   * The shutdown action when ExecutionContext stop
   * former rtc_stopping_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onShutdown(RTC::UniqueId ec_id);

  /***
   * Activate the 3D Shape RTC
   *
   * The activated action (Active state entry action)
   * former rtc_active_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre -
   * @post -
   * 
   */
   virtual RTC::ReturnCode_t onActivated(RTC::UniqueId ec_id);

  /***
   * Deactivate the 3D Shape RTC
   *
   * The deactivated action (Active state exit action)
   * former rtc_active_exit()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre -
   * @post -
   * 
   */
   virtual RTC::ReturnCode_t onDeactivated(RTC::UniqueId ec_id);

  /***
   * Periodic task of the 3D Shape RTC
   *
   * The execution action that is invoked periodically
   * former rtc_active_do()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre -
   * @post -
   * 
   */
   virtual RTC::ReturnCode_t onExecute(RTC::UniqueId ec_id);

  /***
   * onAborting the 3D Shape RTC
   *
   * The aborting action when main logic error occurred.
   * former rtc_aborting_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre -
   * @post -
   * 
   */
   virtual RTC::ReturnCode_t onAborting(RTC::UniqueId ec_id);

  /***
   *
   * The error action in ERROR state
   * former rtc_error_do()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onError(RTC::UniqueId ec_id);

  /***
   * Reset initialization of the 3D Shape RTC
   *
   * The reset action that is invoked resetting
   * This is same but different the former rtc_init_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre -
   * @post -
   * 
   */
   virtual RTC::ReturnCode_t onReset(RTC::UniqueId ec_id);
  
  /***
   *
   * The state update action that is invoked after onExecute() action
   * no corresponding operation exists in OpenRTm-aist-0.2.0
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onStateUpdate(RTC::UniqueId ec_id);

  /***
   *
   * The action that is invoked when execution context's rate is changed
   * no corresponding operation exists in OpenRTm-aist-0.2.0
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onRateChanged(RTC::UniqueId ec_id);


 protected:
  // <rtc-template block="protected_attribute">
  
  // </rtc-template>

  // <rtc-template block="protected_operation">
  
  // </rtc-template>

  // Configuration variable declaration
  // <rtc-template block="config_declare">

  // </rtc-template>

  // DataInPort declaration
  // <rtc-template block="inport_declare">
  RTC::MultiCameraImages m_in_img;
  /*!
   * This port inputs for the 3D shape image data.
   * - Type: RTC::MultiCameraImages
   * - Number: 1
   * - Semantics: -
   * - Unit: -
   * - Frequency: Caused by user actions.
   * - Operation Cycle: two cycles per second.
   */
  InPort<RTC::MultiCameraImages> m_in_imgIn;
  RltObjectIF::ResultObject m_rlt_obj;
  /*!
   * This port inputs for the factorization RTC result.
   * - Type: RltObjectIF::ResultObject
   * - Number: 1
   * - Semantics: -
   * - Unit: -
   * - Frequency: Caused by user actions.
   * - Operation Cycle: two cycles per second.
   */
  InPort<RltObjectIF::ResultObject> m_rlt_objIn;
  
  // </rtc-template>


  // DataOutPort declaration
  // <rtc-template block="outport_declare">
  RTC::CameraImage m_out_img;
  /*!
   * This port outputs for the 3D Shape image data.
   * - Type: RTC::CameraImage
   * - Number: 1
   * - Semantics: -
   * - Unit: -
   * - Frequency: Caused by user actions.
   * - Operation Cycle: two cycles per second.
   */
  OutPort<RTC::CameraImage> m_out_imgOut;
  Rlt3DShapeIF::Result3DShape m_rlt_3d;
  /*!
   * This port outputs the result data of the three dimensional
   * image creation.
   * - Type: Rlt3DShapeIF::Result3DShape
   * - Number: 1
   * - Semantics: -
   * - Unit: -
   * - Frequency: Caused by user actions.
   * - Operation Cycle: two cycles per second.
   */
  OutPort<Rlt3DShapeIF::Result3DShape> m_rlt_3dOut;
  RTC::TimedString m_out_path;
  /*!
   * This port outputs the path information of the 3D shape
   * object.
   * - Type: RTC::TimedString
   * - Number: 1
   * - Semantics: -
   * - Unit: -
   * - Frequency: Caused by user actions.
   * - Operation Cycle: two cycles per second.
   */
  OutPort<RTC::TimedString> m_out_pathOut;
  
  // </rtc-template>

  // CORBA Port declaration
  // <rtc-template block="corbaport_declare">
  /*!
   * This service port notifies completion of 3-D image creation.
   * Interface: Provide the status of 3D image creation.
   */
  RTC::CorbaPort m_Control3DShapeServPort;
  
  // </rtc-template>

  // Service declaration
  // <rtc-template block="service_declare">
  /*!
   * This service port notifies completion of 3-D image creation.
   * - Argument:      3-D image-object creation status
   *                  SHAPE_OF_3D_INITIAL
   *                  SHAPE_OF_3D_CREATING
   *                  SHAPE_OF_3D_CREATED
   * - Return Value:  None.
   * - Exception:     None.
   * - PreCondition:  Exists of the 3-D Shape control RTC, And
   *                  connect with 3-D Shape control RTC.
   * - PostCondition: -
   */
  Creation3DObjResultStatusSVC_impl m_Result3DObjStatusProv;
  
  // </rtc-template>

  // Consumer declaration
  // <rtc-template block="consumer_declare">
  
  // </rtc-template>

 private:
  // <rtc-template block="private_attribute">
  
  // </rtc-template>

  // <rtc-template block="private_operation">
  
  // </rtc-template>

  // add on prototype of function
  void nvdShapeOnInitialize( void );
  void nvdShapeOnActivated( void );
  void nvdShapeOnDeactivated( void );
  void nvdShapeOnAborting( void );
  void nvdShapeOnReset( void );
  void nvdShapeOnExecute( void );
  void nvdCheckofObjectData( void );
  void nvdSetStageControl( ubyte_type );
  void nvdCopy_m_vertex( uword_type );
  void nvdCopy_m_uv( uword_type );
  void nvdCopy_m_face( uword_type );
  void wvdRelease( void );
  void nvdDMemRelease( void );
  void nvdReleaseVertex( void );
  void nvdReleaseUV( void );
  void nvdReleaseFace( void );
  void nvdClrMultipleFlags( void );
  void nvdChkAllocateHeapMem( bool_type );
  void nvdCopyObjectData( uword_type, uword_type, uword_type );
  void nvdCheckofInputImage( void );
  void nvdSaveImageData( RTC::MultiCameraImages * );
  void nvdResultObjectCreation( void );
  void nvdSetResultObjectData( uword_type, uword_type, uword_type );
  void nvdDeleteObjFile( void );
  void nvdDeleteBmpFile( void );
  void nvdCreationDir3d_shape( void );
  void nvdOutStorageDirPath( void );
  void nvdOutCameraImagesCreation( void );
  void _Smooth( void );
  void nvdHeapRelease( void );

  uword_type uw3DShapeOnExecuteStage1( void );
  uword_type uw3DShapeOnExecuteStage2( void );
  uword_type uwSetCameraImages( cv::Mat );

  bool_type bCheckInputObjectData( void );
  bool_type bSaveObjectFile( const char_type * );
  bool_type bSaveResultObjectFile( const char_type * );
  bool_type bAllocateMemory_m_vertex( uword_type );
  bool_type bAllocateMemory_m_uv( uword_type );
  bool_type bAllocateMemory_m_face( uword_type );
  bool_type bCheckMultiCameraImages( RTC::MultiCameraImages );
  bool_type bCheckMultiCameraImagesOfSize( uword_type, uword_type );
  bool_type bMoveObjectFile( void );
  const bool_type _Malloc( void );
  const bool_type bFittingObject( const float& power );

  // Dynamic memory related
  OBJ_MESH::Vertex *mp_vertex;
  OBJ_MESH::UV     *mp_uv;
  OBJ_MESH::Face   *mp_face;

  OBJ_MESH::Vertex **_buf_vertex;
  OBJ_MESH::UV     **_buf_uv;

  bool_type m_allocate_vertex;
  bool_type m_allocate_uv;
  bool_type m_allocate_face;
  bool_type m_allocate_obj;

  uword_type m_allocate_vertex_num;
  uword_type m_allocate_uv_num;
  uword_type m_allocate_face_num;

  // working memory related
  CDP::IMG ref;

  uword_type m_vertex_idx;
  uword_type m_vertex_num;

  uword_type m_uv_idx;
  uword_type m_uv_num;

  uword_type m_face_idx;
  uword_type m_face_num;

  float_type m_face_size;
  float_type m_depth_max;
  float_type m_depth_min;

  IplImage m_IplImage;
  bool_type m_ImageFlag;
  bool_type m_ObjectFlag;

  ubyte_type m_ubStageCtrl;
  ubyte_type m_ubStageCtrlBackUp;

  int IMG_SIZE_WIDTH  = 72;
  int IMG_SIZE_HEIGHT = 54;

  std::string m_strDirPath;

  char_type g_absolute_dir[DEF_MAX_LENGTH];
  uword_type uwAbsolute_dir_err;
};


extern "C"
{
  DLL_EXPORT void rtc_3DShapeInit(RTC::Manager* manager);
};

#endif // RTC_3DSHAPE_H
