// -*- C++ -*-
/*!
 * @file  rtc_3DShape.cpp
 * @brief Processing of three dimensional image creation
 * @date $Date$
 *
 * @author EAST JAPAN ACCOUNTING CENTER CO.,LTD.
 *
 * BSD 3-clause "New" or "Revised" License
 *
 * $Id$
 */

#include "rtc_3DShape.h"

// Module specification
// <rtc-template block="module_spec">
static const char* rtc_3dshape_spec[] =
  {
    "implementation_id", "rtc_3DShape",
    "type_name",         "rtc_3DShape",
    "description",       "Processing of three dimensional image creation",
    "version",           "1.0",
    "vendor",            "EAC",
    "category",          "Image",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    ""
  };
// </rtc-template>

/*!
 * @brief constructor
 * @param manager Maneger Object
 */
rtc_3DShape::rtc_3DShape(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_in_imgIn("ImgIn", m_in_img),
    m_rlt_objIn("Result_factorization", m_rlt_obj),
    m_out_imgOut("ImgOut", m_out_img),
    m_rlt_3dOut("Result_3DShape", m_rlt_3d),
    m_out_pathOut("Storage_dir_path", m_out_path),
    m_Control3DShapeServPort("Control3DShapeServ")

    // </rtc-template>
{
}

/*!
 * @brief destructor
 */
rtc_3DShape::~rtc_3DShape()
{
  wvdRelease();
}


/*!
 * Initialization of the 3D Shape RTC
 */
RTC::ReturnCode_t rtc_3DShape::onInitialize()
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  addInPort("ImgIn", m_in_imgIn);
  addInPort("Result_factorization", m_rlt_objIn);
  
  // Set OutPort buffer
  addOutPort("ImgOut", m_out_imgOut);
  addOutPort("Result_3DShape", m_rlt_3dOut);
  addOutPort("Storage_dir_path", m_out_pathOut);
  
  // Set service provider to Ports
  m_Control3DShapeServPort.registerProvider("Creation3DObjResultStatus", "Creation3DObjResultStatusServ::Creation3DObjResultStatus", m_Result3DObjStatusProv);
  
  // Set service consumers to Ports
  
  // Set CORBA Service Ports
  addPort(m_Control3DShapeServPort);
  
  // </rtc-template>

  // <rtc-template block="bind_config">
  // </rtc-template>
  
  // add to code
  nvdShapeOnInitialize();

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t rtc_3DShape::onFinalize()
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t rtc_3DShape::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t rtc_3DShape::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*!
 * Activate the 3D Shape RTC
 */

RTC::ReturnCode_t rtc_3DShape::onActivated(RTC::UniqueId ec_id)
{
  // add to code
  nvdShapeOnActivated();

  return RTC::RTC_OK;
}

/*!
 * Deactivate the 3D Shape RTC
 */

RTC::ReturnCode_t rtc_3DShape::onDeactivated(RTC::UniqueId ec_id)
{
  // add to code
  std::cout << _ON_DEACTIVATED << std::endl;
  nvdShapeOnDeactivated();

  return RTC::RTC_OK;
}

/*!
 * Periodic task of the 3D Shape RTC
 */

RTC::ReturnCode_t rtc_3DShape::onExecute(RTC::UniqueId ec_id)
{
  // add to code
  nvdShapeOnExecute();

  return RTC::RTC_OK;
}

/*!
 * onAborting the 3D Shape RTC
 */

RTC::ReturnCode_t rtc_3DShape::onAborting(RTC::UniqueId ec_id)
{
  // add to code
  nvdShapeOnAborting();

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t rtc_3DShape::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*!
 * Reset initialization of the 3D Shape RTC
 */

RTC::ReturnCode_t rtc_3DShape::onReset(RTC::UniqueId ec_id)
{
  // add to code start
  nvdShapeOnReset();

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t rtc_3DShape::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t rtc_3DShape::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/***************************************************************************************************
 *  Function        : nvdShapeOnInitialize
 *  Description     : When onInitialize, this function execute the initialization.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DShape::nvdShapeOnInitialize( void )
{
  std::cout << _ON_INITIALIZE << std::endl;

  DWORD result = 0;

  char_type strPath[DEF_MAX_LENGTH];
  char_type strDrive[DEF_MIN_LENGTH];
  char_type strDir[DEF_MAX_LENGTH];
  char_type strFile[DEF_MID_LENGTH];
  char_type strExt[DEF_MIN_LENGTH];

  memset(strPath, 0x00, sizeof(strPath));
  memset(strDrive, 0x00, sizeof(strDrive));
  memset(strDir, 0x00, sizeof(strDir));
  memset(strFile, 0x00, sizeof(strFile));
  memset(strExt, 0x00, sizeof(strExt));
  memset(g_absolute_dir, 0x00, sizeof(g_absolute_dir));

  uwAbsolute_dir_err = DEF_GET_PATH_NG;
  result = GetModuleFileName(NULL, strPath, sizeof(strPath));
  if( result != DEF_GET_PATH_NG )
  {
    uwAbsolute_dir_err = DEF_GET_PATH_OK;
    _splitpath(strPath, strDrive, strDir, strFile, strExt);
    sprintf(g_absolute_dir, "%s%s%s", strDrive, strDir, "3d_shape");
    std::cout << "absolute dir path: " << g_absolute_dir << std::endl;
  }

  m_allocate_vertex = FALSE;
  m_allocate_uv     = FALSE;
  m_allocate_face   = FALSE;
  m_allocate_obj    = FALSE;
  m_allocate_face_num = DEF_NUM_INIT;

}

/***************************************************************************************************
 *  Function        : nvdShapeOnActivated
 *  Description     : When onActivated, this function execute the initialization.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DShape::nvdShapeOnActivated( void )
{
  char_type cbuffer[DEF_MAX_LENGTH];
  char_type cbuf[DEF_MAX_LENGTH];

  memset(cbuffer, 0x00, sizeof(cbuffer));
  memset(cbuf, 0x00, sizeof(cbuf));

  std::cout << _ON_ACTIVATED << std::endl;

  m_allocate_vertex = FALSE;
  m_allocate_uv     = FALSE;
  m_allocate_face   = FALSE;
  m_allocate_obj    = FALSE;
  m_allocate_face_num = DEF_NUM_INIT;

  m_Result3DObjStatusProv.ReqCreation3DObjResultStatus( SHAPE_OF_3D_INITIAL );
  nvdSetStageControl( _3DSHAPE_STAGE1 );

  m_in_img.images.length( IMAGES_LENGTH_2 );
  ref.Create( XIMG_320, YIMG_240 );
  nvdClrMultipleFlags();

  if( uwAbsolute_dir_err == DEF_GET_PATH_OK )
  {
    m_strDirPath = g_absolute_dir;
  }
  else
  {
    GetFullPathName("3d_shape", DEF_MAX_LENGTH, cbuffer, NULL);
    m_strDirPath = cbuffer;
  }

  strcpy( cbuf, m_strDirPath.c_str() );

  if( !PathIsDirectory( cbuf ) )
  {
    (void)_mkdir("3d_shape");
  }
}

/***************************************************************************************************
 *  Function        : nvdShapeOnDeactivated
 *  Description     : When onDeactivated, this function execute the initialization.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DShape::nvdShapeOnDeactivated( void )
{
  m_Result3DObjStatusProv.ReqCreation3DObjResultStatus( SHAPE_OF_3D_INITIAL );
  nvdSetStageControl( _3DSHAPE_STAGE_INIT );

  m_in_img.images.release();
  ref.Release();
  wvdRelease();
  nvdClrMultipleFlags();
}

/***************************************************************************************************
 *  Function        : nvdShapeOnAborting
 *  Description     : When onAborting, this function execute the initialization.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DShape::nvdShapeOnAborting( void )
{
  std::cout << _ON_ABORTING << std::endl;
  std::cout << _ERROR_RESET << std::endl;

  nvdShapeOnDeactivated();
}

/***************************************************************************************************
 *  Function        : nvdShapeOnReset
 *  Description     : When onReset, this function execute the initialization.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DShape::nvdShapeOnReset( void )
{
  std::cout << _ON_RESET << std::endl;
}

/***************************************************************************************************
 *  Function        : nvdShapeOnExecute
 *  Description     : This function execute the main process of 3D shape RTC.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DShape::nvdShapeOnExecute( void )
{
  uword_type auwReturn = RESULT_NG;

  switch ( m_ubStageCtrl )
  {
    case _3DSHAPE_STAGE1:
      auwReturn = uw3DShapeOnExecuteStage1();
      if( auwReturn == RESULT_OK )
      {
        nvdSetStageControl( _3DSHAPE_STAGE2 );
        m_Result3DObjStatusProv.ReqCreation3DObjResultStatus( SHAPE_OF_3D_CREATING );
      }
      break;
    //---------------------------------------------------------------------------------

    //---------------------------------------------------------------------------------
    case _3DSHAPE_STAGE2:
      auwReturn = uw3DShapeOnExecuteStage2();
      if( auwReturn == RESULT_OK )
      {
        std::cout << _3D_SHAPE_SUCCESS << std::endl;
        nvdSetStageControl( _3DSHAPE_STAGE1 );
        m_Result3DObjStatusProv.ReqCreation3DObjResultStatus( SHAPE_OF_3D_CREATED );
      }
      break;
    //---------------------------------------------------------------------------------

    //---------------------------------------------------------------------------------
    case ERROR_DETECTED:
      if( m_ubStageCtrlBackUp != ERROR_DETECTED )
      {
        std::cout << _3D_SHAPE_FAILED << std::endl;
        m_Result3DObjStatusProv.ReqCreation3DObjResultStatus( SHAPE_OF_3D_ERROR );
      }

      break;
    //---------------------------------------------------------------------------------

    default:
      std::cout << _STAGE_ERROR << std::endl;
      break;
  }

  m_ubStageCtrlBackUp = m_ubStageCtrl;
}

/***************************************************************************************************
 *  Function        : nvdSetStageControl
 *  Description     : This function sets the member variable of m_ubStageCtrl.
 *  Argument        : m_ubStageCtrl
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DShape::nvdSetStageControl( ubyte_type a_ubStageCtrl )
{
  if( (a_ubStageCtrl >= _3DSHAPE_STAGE_MIN) && (a_ubStageCtrl < _3DSHAPE_STAGE_MAX) )
  {
    m_ubStageCtrl = a_ubStageCtrl;
  }
  else
  {
    m_ubStageCtrl = _3DSHAPE_STAGE1;
  }
}

/***************************************************************************************************
 *  Function        : uw3DShapeOnExecuteStage1
 *  Description     : This function execute the main stage1 process of 3D shape RTC.
 *  Argument        : nothing special
 *  Return          : information of normal end or abnormal end.
 **************************************************************************************************/
uword_type rtc_3DShape::uw3DShapeOnExecuteStage1( void )
{
  uword_type auwReturn = RESULT_NG;

  nvdCheckofInputImage();
  nvdCheckofObjectData();
  if( (m_ImageFlag == TRUE) && (m_ObjectFlag == TRUE) )
  {
    nvdClrMultipleFlags();
    auwReturn = RESULT_OK;
  }

  return auwReturn;
}

/***************************************************************************************************
 *  Function        : uw3DShapeOnExecuteStage2
 *  Description     : This function execute the main stage2 process of 3D shape RTC.
 *  Argument        : nothing special
 *  Return          : information of normal end or abnormal end.
 **************************************************************************************************/
uword_type rtc_3DShape::uw3DShapeOnExecuteStage2( void )
{
  uword_type auwReturn = RESULT_NG;

  if( !_Malloc() )
  {
    return auwReturn;
  }

  /*************************************************************************************************
  *************************************************************************************************/

  if( bFittingObject( DEPTH_POWER ) )
  {
    if( bSaveResultObjectFile( "3d_shape" ) )
    {
      if( bMoveObjectFile() )
      {
        nvdOutStorageDirPath();
      }
    }
    else
    {
      std::cout << _ERROR_OPEN_FILE << std::endl;
    }

    nvdResultObjectCreation();
    nvdClrMultipleFlags();
    wvdRelease();

    auwReturn = RESULT_OK;
  }

  return auwReturn;
}

/***************************************************************************************************
 *  Function        : bFittingObject
 *  Description     : This function creates Z coordinate data.
 *  Argument        : nothing special
 *  Return          : information of normal end or abnormal end.
 **************************************************************************************************/
const bool_type rtc_3DShape::bFittingObject( const float& power )
{
  uword_type _height = static_cast<uword_type>(IMG_SIZE_HEIGHT);
  uword_type _width  = static_cast<uword_type>(IMG_SIZE_WIDTH);

  if( (m_allocate_uv_num != m_allocate_vertex_num) || m_allocate_uv_num < 0 )
  {
    std::cout << _ERROR_HEAP_MEM_OBJ << std::endl;
    nvdSetStageControl( ERROR_DETECTED );
    return FALSE;
  }

  for( uword_type y = 0; y < _height; ++y )
  {
    for( uword_type x =0; x < _width; ++x )
    {
      _buf_vertex[y][x].z = 0.0f;
    }
  }

  float_type max = 0.0f;
  float_type min = 0.0f;
  uword_type x = 0;
  uword_type y = 0;

  for( uword_type i = 0; i < m_allocate_uv_num; ++i )
  {
    y = static_cast<uword_type>(mp_uv[i].v * _height);
    x = static_cast<uword_type>(mp_uv[i].u * _width);

    if( y < _height || x < _width )
    {
      _buf_vertex[y][x].z = mp_vertex[i].z;
    }
  }

  if( m_depth_min < 0 )
  {
    min = 0;
    max = m_depth_max + fabs( m_depth_min );
  }
  else
  {
    min = m_depth_min;
    max = m_depth_max;
  }

  for( uword_type y = 0; y < _height; ++y )
  {
    for( uword_type x = 0; x < _width; ++x )
    {
      _buf_vertex[y][x].z =  power * ( _buf_vertex[y][x].z / ( max - min ) );
    }
  }

  _Smooth();

  return TRUE;
}

/***************************************************************************************************
 *  Function        : _Smooth
 *  Description     : This function removes noise.
 *  Argument        : nothing special
 *  Argument        : nothing special
 **************************************************************************************************/
void rtc_3DShape::_Smooth( void )
{
  uword_type _height = static_cast<uword_type>(IMG_SIZE_HEIGHT);
  uword_type _width  = static_cast<uword_type>(IMG_SIZE_WIDTH);

  for( uword_type y = 1; y < (_height - 1); ++y )
  {
    for( uword_type x = 1; x < (_width -1); ++x )
    {
      _buf_vertex[y][x].z = static_cast<float_type>(
        _buf_vertex[y+1][x-1].z * 0.0625 + 
        _buf_vertex[y+1][x  ].z * 0.125 + 
        _buf_vertex[y+1][x+1].z * 0.0625 + 
        _buf_vertex[y  ][x-1].z * 0.125 + 
        _buf_vertex[y  ][x  ].z * 0.25 + 
        _buf_vertex[y  ][x+1].z * 0.125 + 
        _buf_vertex[y-1][x-1].z * 0.0625 + 
        _buf_vertex[y-1][x  ].z * 0.125 + 
        _buf_vertex[y-1][x+1].z * 0.0625);
    }
  }

  for( uword_type x = 0; x < _width; ++x )
  {
    _buf_vertex[_height-1][x].z = static_cast<float_type>(_buf_vertex[_height-2][x].z * 0.6 + _buf_vertex[_height-1][x].z * 0.4);
    _buf_vertex[0][x].z = static_cast<float_type>(_buf_vertex[0][x].z * 0.4 + _buf_vertex[1][x].z * 0.6);
  }

  for( uword_type y = 0; y < _height; ++y )
  {
    _buf_vertex[y][0].z = static_cast<float_type>(_buf_vertex[y][0].z * 0.4 + _buf_vertex[y][1].z * 0.6);
    _buf_vertex[y][_width-1].z = static_cast<float_type>(_buf_vertex[y][_width-1].z * 0.4 + _buf_vertex[y][_width-2].z * 0.6);
  }
}

/***************************************************************************************************
 *  Function        : _Malloc
 *  Description     : This function allocate dynamic memory.
 *  Argument        : nothing special
 *  Return          : information of normal end or abnormal end.
 **************************************************************************************************/
const bool_type rtc_3DShape::_Malloc( void )
{
  uword_type _height = static_cast<uword_type>(IMG_SIZE_HEIGHT);
  uword_type _width  = static_cast<uword_type>(IMG_SIZE_WIDTH);

  try
  {
    _buf_vertex = new OBJ_MESH::Vertex*[ _height ];
    _buf_uv     = new OBJ_MESH::UV*[ _height ];

    for( uword_type i = 0; i < _height; i++ )
    {
      _buf_vertex[i] = new OBJ_MESH::Vertex[ _width ];
      _buf_uv[i]     = new OBJ_MESH::UV[ _width ];
    }
  }
  catch( std::bad_alloc )
  {
    std::cout << _ERROR_ALLOCATION << std::endl;
    nvdSetStageControl( ERROR_DETECTED );
    m_allocate_obj = FALSE;
    return FALSE;
  }

  for( uword_type y = 0; y < _height; ++y )
  {
    for( uword_type x =0; x < _width; ++x )
    {
      _buf_vertex[y][x].y = static_cast<float_type>(y);
      _buf_vertex[y][x].x = static_cast<float_type>(x);
      _buf_vertex[y][x].z = 0.0f;

      _buf_uv[y][x].v = y * ( 1.0f / (_height-1) );
      _buf_uv[y][x].u = x * ( 1.0f / (_width-1) );
    }
  }

  m_allocate_obj = TRUE;

  return TRUE;
}

/***************************************************************************************************
 *  Function        : bMoveObjectFile
 *  Description     : This function moves the object file.
 *  Argument        : object file
 *  Return          : nothing special
 **************************************************************************************************/
bool_type rtc_3DShape::bMoveObjectFile( void )
{
  bool_type abReturn = FALSE;

  nvdDeleteObjFile();
  nvdDeleteBmpFile();
  nvdCreationDir3d_shape();

  if( MoveFile("3d_shape.ftv", "./3d_shape/3d_shape.ftv") )
  {
    if( MoveFile("3d_shape.bmp", "./3d_shape/3d_shape.bmp") )
    {
      abReturn = TRUE;
    }
  }

  return abReturn;
}

/***************************************************************************************************
 *  Function        : nvdDeleteObjFile
 *  Description     : This function deletes the object file if an object file exists.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DShape::nvdDeleteObjFile( void )
{
  if( PathFileExists( "./3d_shape/3d_shape.ftv" ) )
  {
    (void)DeleteFile("./3d_shape/3d_shape.ftv");
  }
}

/***************************************************************************************************
 *  Function        : nvdDeleteBmpFile
 *  Description     : This function deletes the bmp file if an bmp file exists.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DShape::nvdDeleteBmpFile( void )
{
  if( PathFileExists( "./3d_shape/3d_shape.bmp" ) )
  {
    (void)DeleteFile("./3d_shape/3d_shape.bmp");
  }
}

/***************************************************************************************************
 *  Function        : nvdCreationDir3d_shape
 *  Description     : This function creates a 3d-shape directory when there is no 3d-shape directory.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DShape::nvdCreationDir3d_shape( void )
{
  char_type cbuf[DEF_MAX_LENGTH];
  memset(cbuf, 0x00, sizeof(cbuf));

  strcpy( cbuf, m_strDirPath.c_str() );
  if( !PathIsDirectory( cbuf ) )
  {
    (void)_mkdir("3d_shape");
  }
}

/***************************************************************************************************
 *  Function        : nvdOutStorageDirPath
 *  Description     : This function outputs the path information of the object file.
 *  Argument        : Object file storage information
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DShape::nvdOutStorageDirPath( void )
{
  m_out_path.data = CORBA::string_dup( m_strDirPath.c_str() );
  bool_type abWrtResult = m_out_pathOut.write();
  if( abWrtResult == FALSE )
  {
    std::cout << _ERROR_WRITE_PATH << std::endl;
  }
}

/***************************************************************************************************
 *  Function        : nvdCheckofInputImage
 *  Description     : This function checks the inputs image data.
 *  Argument        : Input port of image data
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DShape::nvdCheckofInputImage( void )
{
  if( m_in_imgIn.isNew() )
  {
    m_ImageFlag = FALSE;
    m_in_imgIn.read();
    if( bCheckMultiCameraImages(m_in_img) )
    {
      nvdSaveImageData( &m_in_img );
      nvdOutCameraImagesCreation();
    }
  }
}

/***************************************************************************************************
 *  Function        : nvdCheckofObjectData
 *  Description     : This function checks the inputs object data.
 *  Argument        : Input port of object data
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DShape::nvdCheckofObjectData( void )
{
  if( bCheckInputObjectData() )
  {
    m_ObjectFlag = FALSE;
    if( bSaveObjectFile("rlt_factorization.obj") )
    {
      m_ObjectFlag = TRUE;
    }
  }
}

/***************************************************************************************************
 *  Function        : bCheckMultiCameraImages
 *  Description     : This function will check whether the mult-cameraimages data is valid.
 *  Argument        : multi-cameraimages
 *  Return          : information of normal end or abnormal end
 **************************************************************************************************/
bool_type rtc_3DShape::bCheckMultiCameraImages( RTC::MultiCameraImages MultiImage )
{
  bool_type abReturn = FALSE;

  uword_type imgwidth;
  uword_type imgheight;
  ubyte_type result[2];

  ubyte_type i = 0;
  for( i = 0; i < IMAGES_LENGTH_2; i++ )
  {
    imgwidth  = MultiImage.images[i].width;
    imgheight = MultiImage.images[i].height;

    result[i] = FALSE;
    if( bCheckMultiCameraImagesOfSize(imgwidth, imgheight) )
    {
      if( MultiImage.images[i].bpp == FORMAT_COLOR24 )
      {
        result[i] = TRUE;
      }
      else
      {
        std::cout << _ERROR_IMAGE_COLOR << std::endl;
      }
    }
  }

  if( (result[0] == TRUE) && (result[1] == TRUE) )
  {
    abReturn = TRUE;
  }

  return abReturn;
}

/***************************************************************************************************
 *  Function        : bCheckMultiCameraImagesOfSize
 *  Description     : This function will check the range of the cameraimage data.
 *  Argument        : cameraimage
 *  Return          : information of normal end or abnormal end
 **************************************************************************************************/
bool_type rtc_3DShape::bCheckMultiCameraImagesOfSize( uword_type imgwidth, uword_type imgheight )
{
  bool_type abReturn = FALSE;

  if( (imgwidth == 0) || (imgheight == 0) )
  {
    std::cout << _ERROR_IMAGE_EMPTY << std::endl;
  }
  else if( (imgwidth <= XIMG_320) && (imgheight <= YIMG_240) )
  {
    abReturn = TRUE;
  }
  else
  {
    std::cout << _ERROR_IMAGE_SIZE << std::endl;
  }

  return abReturn;
}

/***************************************************************************************************
 *  Function        : nvdSaveImageData
 *  Description     : This function saves image data.
 *  Argument        : *image_p
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DShape::nvdSaveImageData( RTC::MultiCameraImages *image_p )
{
  cv::Mat m_image;

  m_image = cv::Mat( cv::Size( image_p->images[IMG_INP_C].width, image_p->images[IMG_INP_C].height ), CV_8UC3 );
  memcpy( m_image.data, (void *)&(image_p->images[IMG_INP_C].pixels[0]), image_p->images[IMG_INP_C].pixels.length() );
  m_IplImage = m_image;

  cvResize( &m_IplImage, ref.iplImage(), CDP_IMAGE_RESIZE_TYPE );
  m_ImageFlag = TRUE;
}

/***************************************************************************************************
 *  Function        : bCheckInputObjectData
 *  Description     : This function allocates dynamic memory and saves object data.
 *  Argument        : factorization RTC object data
 *  Return          : information of normal end or abnormal end
 **************************************************************************************************/
bool_type rtc_3DShape::bCheckInputObjectData( void )
{
  bool_type abReturn = FALSE;

  if( m_rlt_objIn.isNew() )
  {
    m_rlt_objIn.read();
    if( bAllocateMemory_m_vertex( m_rlt_obj.vertex_idx_len ) )
    {
      if( bAllocateMemory_m_uv( m_rlt_obj.uv_idx_len ) )
      {
        if( bAllocateMemory_m_face( m_rlt_obj.face_idx_len ) )
        {
          nvdCopyObjectData( m_rlt_obj.vertex_idx_len, m_rlt_obj.uv_idx_len, m_rlt_obj.face_idx_len );
          abReturn = TRUE;
        }
      }
    }
    nvdChkAllocateHeapMem( abReturn );
  }

  return abReturn;
}

/***************************************************************************************************
 *  Function        : bAllocateMemory_m_vertex
 *  Description     : This function allocates the area of the m_vertex variable.
 *  Argument        : dynamic memory length
 *  Return          : information of normal end or abnormal end
 **************************************************************************************************/
bool_type rtc_3DShape::bAllocateMemory_m_vertex( uword_type len )
{
  m_allocate_vertex = TRUE;

  try
  {
    mp_vertex = new OBJ_MESH::Vertex[ len ];
    m_allocate_vertex_num = len;
  }
  catch( std::bad_alloc )
  {
    m_allocate_vertex = FALSE;
  }

  return m_allocate_vertex;
}

/***************************************************************************************************
 *  Function        : bAllocateMemory_m_uv
 *  Description     : This function allocates the area of the m_uv variable.
 *  Argument        : dynamic memory length
 *  Return          : information of normal end or abnormal end
 **************************************************************************************************/
bool_type rtc_3DShape::bAllocateMemory_m_uv( uword_type len )
{
  m_allocate_uv = TRUE;

  try
  {
    mp_uv = new OBJ_MESH::UV[ len ];
    m_allocate_uv_num = len;
  }
  catch( std::bad_alloc )
  {
    m_allocate_uv = FALSE;
  }

  return m_allocate_uv;
}

/***************************************************************************************************
 *  Function        : bAllocateMemory_m_face
 *  Description     : This function allocates the area of the m_face variable.
 *  Argument        : dynamic memory length
 *  Return          : information of normal end or abnormal end
 **************************************************************************************************/
bool_type rtc_3DShape::bAllocateMemory_m_face( uword_type len )
{
  m_allocate_face = TRUE;

  try
  {
    mp_face = new OBJ_MESH::Face[ len ];
    m_allocate_face_num = len;
  }
  catch( std::bad_alloc )
  {
    m_allocate_face = FALSE;
    m_allocate_face_num = DEF_NUM_INIT;
  }

  uword_type i = 0;
  for ( i = 0; i < len; i++ )
  {
    try
    {
      mp_face[i].vertex_index = new int [4];
    }
    catch( std::bad_alloc )
    {
      m_allocate_face = FALSE;
      m_allocate_face_num = DEF_NUM_INIT;
    }
  }

  return m_allocate_face;
}

/***************************************************************************************************
 *  Function        : nvdCopyObjectData
 *  Description     : This function calls a function that copies object data.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DShape::nvdCopyObjectData( uword_type vertex_len, uword_type uv_len, uword_type face_len )
{
  nvdCopy_m_vertex( vertex_len );
  nvdCopy_m_uv( uv_len );
  nvdCopy_m_face( face_len );
}

/***************************************************************************************************
 *  Function        : nvdCopy_m_vertex
 *  Description     : This function copies data from "m_rlt_obj.p_vertex" to "m_vertex".
 *  Argument        : length
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DShape::nvdCopy_m_vertex( uword_type len )
{
  m_vertex_idx = len;
  m_rlt_obj.p_vertex.length(len);

  m_vertex_num = m_rlt_obj.vertex_num_len;

  uword_type i = 0;
  for (i = 0; i < len; i++)
  {
    mp_vertex[i].x = m_rlt_obj.p_vertex[i].x;
    mp_vertex[i].y = m_rlt_obj.p_vertex[i].y;
    mp_vertex[i].z = m_rlt_obj.p_vertex[i].z;
  }

  m_rlt_obj.p_vertex.release();
}

/***************************************************************************************************
 *  Function        : nvdCopy_m_uv
 *  Description     : This function copies data from "m_rlt_obj.p_uv" to "m_uv".
 *  Argument        : length
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DShape::nvdCopy_m_uv( uword_type len )
{
  m_uv_idx = len;
  m_rlt_obj.p_uv.length(len);

  m_uv_num = m_rlt_obj.uv_num_len;

  uword_type i = 0;
  for (i = 0; i < len; i++)
  {
    mp_uv[i].u = m_rlt_obj.p_uv[i].u;
    mp_uv[i].v = m_rlt_obj.p_uv[i].v;
  }

  m_rlt_obj.p_uv.release();
}

/***************************************************************************************************
 *  Function        : nvdCopy_m_face
 *  Description     : This function copies data from "m_rlt_obj.p_face" to "m_face".
 *  Argument        : length
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DShape::nvdCopy_m_face( uword_type len )
{
  uword_type auw_element = 0;

  m_face_idx = len;
  m_rlt_obj.p_face.length(len);

  m_face_num  = m_rlt_obj.face_num_len;
  m_face_size = m_rlt_obj.face_size;
  m_depth_max = m_rlt_obj.depth_max;
  m_depth_min = m_rlt_obj.depth_min;

  uword_type i = 0;
  for ( i = 0; i < len; i++ )
  {
    mp_face[i].type    = m_rlt_obj.p_face[i].type;
    mp_face[i].element = m_rlt_obj.p_face[i].element;

    auw_element = static_cast<uword_type>(mp_face[i].element);
    uword_type j = 0;
    for( j = 0; j < auw_element; j++ )
    {
      mp_face[i].vertex_index[j] = m_rlt_obj.p_face[i].vertex_idx[j];
    }
  }

  m_rlt_obj.p_face.release();
}

/***************************************************************************************************
 *  Function        : nvdChkAllocateHeapMem
 *  Description     : This function updates the stage information on error.
 *  Argument        : memory allocation error information
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DShape::nvdChkAllocateHeapMem( bool_type abJudgeFlg )
{
  if( !abJudgeFlg )
  {
    std::cout << _ERROR_ALLOCATION << std::endl;
    std::cout << _ERROR_HEAP_MEM << std::endl;
    nvdSetStageControl( ERROR_DETECTED );
  }
}

/***************************************************************************************************
 *  Function        : bSaveObjectFile
 *  Description     : This function saves the factorization-RTC object data to a file.
 *  Argument        : factorization RTC object data
 *  Return          : information of normal end or abnormal end
 **************************************************************************************************/
bool_type rtc_3DShape::bSaveObjectFile( const char_type *filename )
{
  FILE *fp;
  uword_type auw_element = 0;

  if( filename == NULL ){
    return false;
  }

  fp = fopen( filename, "wb" );
  if( fp == NULL ){
    return false;
  }

  fprintf(fp, "# Created by AWOFormatter\n\n");
  fprintf(fp, "mtllib c1_0.mtl\n\n");

  for( uword_type i = 0; i < m_vertex_idx; i++ )
    fprintf( fp, "v %f %f %f\n", mp_vertex[i].x, mp_vertex[i].y, mp_vertex[i].z );

  fprintf( fp, "# %d vertices\n\n", m_vertex_idx );

  for( uword_type i = 0; i < m_uv_idx; i++ )
    fprintf( fp, "vt %f %f\n", mp_uv[i].u, mp_uv[i].v );
  fprintf( fp, "# %d vertices\n\n", m_uv_idx );

  fprintf( fp, "g obj1\nusemtl mat1\n" );
  for( uword_type i = 0; i < m_face_idx; i++ ){
    fprintf( fp, "f" );
    auw_element = static_cast<uword_type>(mp_face[i].element);
    for( uword_type j = 0; j < auw_element; j++ ){
      fprintf( fp, " %d/%d", (mp_face[i].vertex_index[j] + 1), (mp_face[i].vertex_index[j] + 1) );
    }
    fprintf( fp, "\n" );
  }

  fclose( fp );

  return true;
}

/***************************************************************************************************
 *  Function        : bSaveResultObjectFile
 *  Description     : This function saves the object data of 3D-shape in a file.
 *  Argument        : 3D-shape RTC object data
 *  Return          : information of normal end or abnormal end.
 **************************************************************************************************/
bool_type rtc_3DShape::bSaveResultObjectFile( const char_type *filename )
{
  FILE *fp;
  char_type buf[1024];
  IplImage *_texture;

  if( filename == NULL ){
    return false;
  }

  sprintf( buf, "%s.ftv", filename );

  fp = fopen( buf, "w" );
  if( fp == NULL ){
    return false;
  }

  fprintf( fp, "# Created by 3D shape file\n\n" );

  _texture = cvCloneImage( ref.iplImage() );

  if( _texture != NULL ){
    sprintf( buf, "%s.bmp", filename );
    fprintf( fp, "T %s\n\n", buf );
    cvSaveImage( buf, _texture );
  }
  cvReleaseImage( &_texture );

  fprintf( fp, "S %d %d\n\n", IMG_SIZE_WIDTH, IMG_SIZE_HEIGHT );

  uword_type _height = static_cast<uword_type>(IMG_SIZE_HEIGHT);
  uword_type _width  = static_cast<uword_type>(IMG_SIZE_WIDTH);

  for( uword_type y = 0; y < _height; ++y ){
    for( uword_type x = 0; x < _width; ++x ){
      fprintf( fp, "d %d %d %lf\n", y, x, _buf_vertex[y][x].z );
    }
  }

  fprintf( fp, "#%u vertices\n\n" , (IMG_SIZE_HEIGHT * IMG_SIZE_WIDTH) );

  for( uword_type y = 0; y < _height; ++y ){
    for( uword_type x = 0; x < _width; ++x ){
      fprintf( fp, "u %d %d %lf\n", y, x, _buf_uv[y][x].u );
    }
  }

  fprintf( fp, "#%u texture-u\n\n" , (IMG_SIZE_HEIGHT * IMG_SIZE_WIDTH) );

  for( uword_type y = 0; y < _height; ++y ){
    for( uword_type x = 0; x < _width; ++x ){
      fprintf( fp, "v %d %d %lf\n", y, x, _buf_uv[y][x].v );
    }
  }

  fprintf( fp, "#%u texture-v\n\n" , (IMG_SIZE_HEIGHT * IMG_SIZE_WIDTH) );

  fclose( fp );

  return true;
}

/***************************************************************************************************
 *  Function        : nvdResultObjectCreation
 *  Description     : This function sets object data to be output port.
 *  Argument        : object data
 *  Return          : output port
 **************************************************************************************************/
void rtc_3DShape::nvdResultObjectCreation( void )
{
  uword_type _height = static_cast<uword_type>(IMG_SIZE_HEIGHT);
  uword_type _width  = static_cast<uword_type>(IMG_SIZE_WIDTH);
  uword_type len     = (_height * _width);

  m_rlt_3d.p_obj.length( len );

  m_rlt_3d.rows = _height;
  m_rlt_3d.cols = _width;

  nvdSetResultObjectData( _height, _width, len );

  bool_type abWrtResult = m_rlt_3dOut.write();
  if( abWrtResult == FALSE )
  {
    // std::cout << _ERROR_WRITE_OBJ << std::endl;
  }

  /* Initialize dynamic memory */
  m_rlt_3d.p_obj.release();
}

/***************************************************************************************************
 *  Function        : nvdSetResultObjectData
 *  Description     : This function sets the object data to the output port.
 *  Argument        : object data
 *  Return          : output port
 **************************************************************************************************/
void rtc_3DShape::nvdSetResultObjectData( uword_type _height, uword_type _width, uword_type len )
{
  uword_type i = 0;

  for( uword_type y = 0; y < _height; ++y )
  {
    for( uword_type x = 0; x < _width; ++x )
    {
      if( i < len )
      {
        m_rlt_3d.p_obj[i].vertex = _buf_vertex[y][x].z;
        m_rlt_3d.p_obj[i].text_u = _buf_uv[y][x].u;
        m_rlt_3d.p_obj[i].text_v = _buf_uv[y][x].v;
        i++;
      }
    }
  }
}

/***************************************************************************************************
 *  Function        : wvdRelease
 *  Description     : This function releases dynamic memory.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DShape::wvdRelease( void )
{
  nvdDMemRelease();
  nvdHeapRelease();
}

/***************************************************************************************************
 *  Function        : nvdHeapRelease
 *  Description     : This function releases dynamic memory.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DShape::nvdHeapRelease( void )
{
  uword_type _height = static_cast<uword_type>(IMG_SIZE_HEIGHT);
  uword_type _width  = static_cast<uword_type>(IMG_SIZE_WIDTH);

  if( m_allocate_obj == TRUE )
  {
    if( _buf_vertex != NULL && _buf_uv != NULL )
    {
      for( uword_type i = 0; i < _height; ++i )
      {
        delete [] _buf_vertex[i];
        delete [] _buf_uv[i];

        _buf_vertex[i] = NULL;
        _buf_uv[i] = NULL;
      }
      delete [] _buf_uv;
      delete [] _buf_vertex;

      _buf_uv = NULL;
      _buf_vertex = NULL;
    }
  }

}

/***************************************************************************************************
 *  Function        : nvdDMemRelease
 *  Description     : This function releases dynamic memory.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DShape::nvdDMemRelease( void )
{
  nvdReleaseVertex();
  nvdReleaseUV();
  nvdReleaseFace();
}

/***************************************************************************************************
 *  Function        : nvdReleaseVertex
 *  Description     : This function releases dynamic memory of mp_vertex.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DShape::nvdReleaseVertex( void )
{
  if( m_allocate_vertex )
  {
    if( mp_vertex != NULL )
    {
      delete [] mp_vertex;
      mp_vertex = NULL;
      m_allocate_vertex = FALSE;
    }
  }
}

/***************************************************************************************************
 *  Function        : nvdReleaseUV
 *  Description     : This function releases dynamic memory of mp_uv.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DShape::nvdReleaseUV( void )
{
  if( m_allocate_uv )
  {
    if( mp_uv != NULL )
    {
      delete [] mp_uv;
      mp_uv = NULL;
      m_allocate_uv = FALSE;
    }
  }
}

/***************************************************************************************************
 *  Function        : nvdReleaseFace
 *  Description     : This function releases dynamic memory of mp_face.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DShape::nvdReleaseFace( void )
{
  if( m_allocate_face )
  {
    uword_type i = 0;
    for( i = 0; i < m_allocate_face_num; i++ )
    {
      if( mp_face[i].vertex_index != NULL )
      {
        delete [] mp_face[i].vertex_index;
      }
    }

    if( mp_face != NULL )
    {
      delete [] mp_face;
      mp_face = NULL;
      m_allocate_face = FALSE;
      m_allocate_face_num = DEF_NUM_INIT;
    }
  }
}

/***************************************************************************************************
 *  Function        : nvdClrMultipleFlags
 *  Description     : This function clears multiple flags.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DShape::nvdClrMultipleFlags( void )
{
  m_ImageFlag  = FALSE;
  m_ObjectFlag = FALSE;
}

/***************************************************************************************************
 *  Function        : nvdOutCameraImagesCreation
 *  Description     : This function sets the data to the output port.
 *  Argument        : nothing special
 *  Return          : data of output port
 **************************************************************************************************/
void rtc_3DShape::nvdOutCameraImagesCreation( void )
{
  bool_type abWrtResult = TRUE;
  cv::Mat m_image;

  m_image = cv::cvarrToMat( ref.iplImage() );
  if( RESULT_OK == uwSetCameraImages( m_image ) )
  {
    abWrtResult = m_out_imgOut.write();
    if( abWrtResult == FALSE )
    {
      // std::cout << _ERROR_WRITE_IMG << std::endl;
    }
  }
}

/***************************************************************************************************
 *  Function        : uwSetCameraImages
 *  Description     : This function sets the data of the CameraImage type.
 *  Argument        : Mat type dsp_image
 *  Return          : information of normal end or abnormal end.
 **************************************************************************************************/
uword_type rtc_3DShape::uwSetCameraImages( cv::Mat dsp_image )
{
  uword_type auwReturn = RESULT_NG;

  if( dsp_image.empty() )
  {
    std::cout << _ERROR_IMAGE_EMPTY << std::endl;
  }
  else
  {
    sword_type swlength = ( ( dsp_image.channels() * dsp_image.cols ) * dsp_image.rows );
    m_out_img.width  = dsp_image.cols;
    m_out_img.height = dsp_image.rows;
    m_out_img.bpp    = FORMAT_COLOR24;
    m_out_img.format = "bitmap";
    m_out_img.fDiv   = 0;
    m_out_img.pixels.length( swlength );
    ::memcpy( ( void * )&( m_out_img.pixels[ 0 ] ) , dsp_image.data , swlength );

    auwReturn = RESULT_OK;
  }

  return auwReturn;
}

/***************************************************************************************************
 **************************************************************************************************/


extern "C"
{
 
  void rtc_3DShapeInit(RTC::Manager* manager)
  {
    coil::Properties profile(rtc_3dshape_spec);
    manager->registerFactory(profile,
                             RTC::Create<rtc_3DShape>,
                             RTC::Delete<rtc_3DShape>);
  }
  
};


